#include <winsock2.h>
#include <io.h>
#include <errno.h>
#include <assert.h>
#include "sfds.h"

static int winsock_initialized = 0;

#define IS_WINSOCK_INITIALIZED() (winsock_initialized != 0)

void WSHELPinitialize();

static int
getWSAErrno()
{
	int wsaerrno = WSAGetLastError();
	if (wsaerrno == WSAEWOULDBLOCK)
		return EAGAIN;
	if (wsaerrno == WSAEFAULT)
		return EFAULT;
	if (wsaerrno == WSAEINVAL)
		return EINVAL;
	return wsaerrno;
}

int
WSHELPsocket (int af, int type, int protocol)
{
	SOCKET sock;
	int sfd;

	/* Verify that winsock has been initialized */
	if (!IS_WINSOCK_INITIALIZED())
	{
		WSHELPinitialize();
	}

	/* Clear errno */
	errno = 0;

	/* Pass through to base layer */
	sock = socket(af, type, protocol);

	/* Handle errors */
	if (sock == INVALID_SOCKET)
	{
		errno = getWSAErrno();
		return -1;
	}

	/* Map the socket */
	sfd = allocate_sfd((int)sock);

	return sfd;
}

int
WSHELPsetsockopt (int sfd, int level, int optname, const char* optval, int optlen)
{
	SOCKET sock;
	int ret;

	/* Clear errno */
	errno = 0;

	/* Get the SOCKET */
	sock = (SOCKET) sfd_to_real(sfd, NULL);

	/* Call the underlying function */
	ret = setsockopt(sock, level, optname, optval, optlen);

	/* Check for errors */
	if (ret == SOCKET_ERROR)
	{
		errno = getWSAErrno();
		return -1;
	}

	return 0;
}

int
WSHELPioctlsocket(int sfd, long cmd, u_long* argp)
{
	SOCKET sock;
	int ret;

	/* Clear errno */
	errno = 0;

	/* Get the SOCKET */
	sock = (SOCKET) sfd_to_real(sfd, NULL);

	/* Call the underlying function */
	ret = ioctlsocket(sock, cmd, argp);

	/* Check for errors */
	if (ret == SOCKET_ERROR)
	{
		errno = getWSAErrno();
		return -1;
	}

	return 0;
}

int
WSHELPlisten (int sfd, int backlog)
{
	SOCKET sock;
	int ret;

	/* Clear errno */
	errno = 0;

	/* Get the SOCKET */
	sock = (SOCKET) sfd_to_real(sfd, NULL);

	/* Call the underlying function */
	ret = listen(sock, backlog);

	/* Check for errors */
	if (ret == SOCKET_ERROR)
	{
		errno = getWSAErrno();
		return -1;
	}

	return 0;
}

int
WSHELPbind (int sfd, const struct sockaddr *name, int namelen)
{
	SOCKET sock;
	int ret;

	/* Clear errno */
	errno = 0;

	/* Get the SOCKET */
	sock = (SOCKET) sfd_to_real(sfd, NULL);

	/* Call the underlying function */
	ret = bind(sock, name, namelen);

	/* Check for errors */
	if (ret == SOCKET_ERROR)
	{
		errno = getWSAErrno();
		return -1;
	}

	return 0;
}

int
WSHELPconnect (int sfd, const struct sockaddr* name, int namelen)
{
	SOCKET sock;
	int ret;

	/* Clear errno */
	errno = 0;

	/* Get the SOCKET */
	sock = (SOCKET) sfd_to_real(sfd, NULL);

	/* Call the underlying function */
	ret = connect(sock, name, namelen);

	/* Check for errors */
	if (ret == SOCKET_ERROR)
	{
		errno = getWSAErrno();
		return -1;
	}

	return 0;
}

int
WSHELPaccept(int sfd, struct sockaddr* addr, int* addrlen)
{
	SOCKET sock;
	SOCKET new_sock;
	int new_sfd;

	/* Clear errno */
	errno = 0;

	/* Get the SOCKET */
	sock = (SOCKET) sfd_to_real(sfd, NULL);

	/* Call the underlying function */
	new_sock = accept(sock, addr, addrlen);

	/* Handle errors */
	if (new_sock == INVALID_SOCKET)
	{
		errno = getWSAErrno();
		return -1;
	}

	/* Map the socket */
	new_sfd = allocate_sfd((int)new_sock);

	return new_sfd;
}

static void
split_sfdsets(fd_set *sfdset, fd_set *fdset, fd_set *sockset, int *have_fds, int *have_sockets)
{
	unsigned int i;

	FD_ZERO(fdset);
	FD_ZERO(sockset);

	/* Nothing to do if the input list is NULL or zero length */
	if (sfdset == NULL || sfdset->fd_count == 0)
	{
		return;
	}

	/* Split the input array into a set of fds and a set of SOCKETS, indicate if we found any of each */
	for (i = 0; i < sfdset->fd_count; ++i)
	{
		int sfd = (int) sfdset->fd_array[i];

		if (SFD_IS_REAL_FD(sfd))
		{
			*have_fds = 1;

			FD_SET(sfd, fdset);
		}
		else
		{
			*have_sockets = 1;

			FD_SET(sfd_to_real(sfd, NULL), sockset);
		}
	}
}

int
WSHELPselect (int fds, fd_set* readsfds, fd_set* writesfds, fd_set* exceptsfds, const struct timeval *timeout)
{
	int have_sockets = 0;
	int have_fds = 0;
	fd_set readfds;
	fd_set writefds;
	fd_set exceptfds;
	fd_set readsocks;
	fd_set writesocks;
	fd_set exceptsocks;
	const struct timeval zero = {0, 0};
	const struct timeval slice = {0, 0};
	int ret;

	/* Create socket maps for read, write, and except fd_sets */
	split_sfdsets(readsfds, &readfds, &readsocks, &have_fds, &have_sockets);
	split_sfdsets(writesfds, &writefds, &writesocks, &have_fds, &have_sockets);
	split_sfdsets(exceptsfds, &exceptfds, &exceptsocks, &have_fds, &have_sockets);

	/* If we have nothing, allow select () to set the error */
	if (!have_sockets && !have_fds)
	{
		ret = select(0, &readsocks, &writesocks, &exceptsocks, &zero);
		
		if (ret == SOCKET_ERROR)
		{
			errno = WSAGetLastError();
			return -1;
		}
	}

	/* We only have sockets, this is easy */
	if (have_sockets && !have_fds)
	{
		ret = select(0, &readsocks, &writesocks, &exceptsocks, timeout);
		
		/* Bail if select failed for some reason */
		if (ret == SOCKET_ERROR)
		{
			errno = getWSAErrno();
			return -1;
		}

		/* TODO: Loop through each socks fd_set and check for being set, then set it in the sfd fd_set */
		
		/* If timeout, then none are set */
		if (ret == 0)
		{
			if (readsfds) FD_ZERO(readsfds);
			if (writesfds) FD_ZERO(writesfds);
			if (exceptsfds) FD_ZERO(exceptsfds);

			return ret;
		}

		return ret;
	}


	/* TODO: The complex mode: both fds and sockets, yuk, or anything with real fds */
	//DebugBreak();

	return -1;
}


/*
 * IO functions
 */

int
WSHELPread(int sfd, void *dst, unsigned int max)
{
	SOCKET sock;
	int ret;

	/* Handle right away if this is a real fd */
	if (SFD_IS_REAL_FD(sfd))
	{
		return _read(sfd, dst, max);
	}

	/* Clear errno */
	errno = 0;

	/* Get the SOCKET */
	sock = (SOCKET) sfd_to_real(sfd, NULL);

	/* Call the underlying function */
	ret = recv(sock, dst, max, 0);

	/* Check for errors */
	if (ret == SOCKET_ERROR)
	{
		errno = getWSAErrno();
		return -1;
	}

	return ret;
}

int
WSHELPwrite(int sfd, const void *buf, unsigned int max)
{
	SOCKET sock;
	int ret;

	/* Handle right away if this is a real fd */
	if (SFD_IS_REAL_FD(sfd))
	{
		return _write(sfd, buf, max);
	}

	/* Clear errno */
	errno = 0;

	/* Get the SOCKET */
	sock = (SOCKET) sfd_to_real(sfd, NULL);

	/* Call the underlying function */
	ret = send(sock, buf, max, 0);

	/* Check for errors */
	if (ret == SOCKET_ERROR)
	{
		errno = getWSAErrno();
		return -1;
	}

	return ret;
}

int
WSHELPclose(int sfd)
{
	SOCKET sock;
	int ret;

	/* Handle right away if this is a real fd */
	if (SFD_IS_REAL_FD(sfd))
	{
		return _close(sfd);
	}

	/* Clear errno */
	errno = 0;

	/* Get the SOCKET */
	sock = (SOCKET) sfd_to_real(sfd, NULL);

	/* Call the underlying function */
	ret = closesocket(sock);

	/* Remove mapping table entry */
	free_sfd(sfd);

	/* Check for errors */
	if (ret == SOCKET_ERROR)
	{
		errno = getWSAErrno();
		return -1;
	}

	return ret;
}


/*
 * Internal functions
 */

void
WSHELPinitialize()
{
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;
	 
	wVersionRequested = MAKEWORD(2, 2);
	 
	err = WSAStartup(wVersionRequested, &wsaData);
	assert(err == 0);
	 
	/* Confirm that the WinSock DLL supports 2.2.*/
	/* Note that if the DLL supports versions greater    */
	/* than 2.2 in addition to 2.2, it will still return */
	/* 2.2 in wVersion since that is the version we      */
	/* requested.                                        */
	assert(LOBYTE(wsaData.wVersion) == 2 &&
			HIBYTE(wsaData.wVersion) == 2);
	
	/* The WinSock DLL is acceptable. Proceed. */
	winsock_initialized = 1;
}
